"""
Powerwiki models
"""
import os

from django.conf import settings
from django.contrib.auth import get_user_model
from django.contrib.auth.models import AbstractUser
from django.core.validators import RegexValidator
from django.db import models
from django.utils.html import mark_safe

from . import app_settings, constants, utils
from .markup.loader import engine_field_choices, get_engine, load_engines


if "django.contrib.postgres" in settings.INSTALLED_APPS:
    from django.contrib.postgres.search import SearchVector

#  Preload engines
load_engines()


class Wiki(models.Model):
    """
    Wiki
    """

    title = models.CharField(max_length=255, help_text="Title of the wiki",)
    slug = models.SlugField(unique=True, help_text="Slug for the wiki",)
    description = models.TextField(blank=True, help_text="Description of wiki")
    perm_read = models.IntegerField(
        default=constants.PERM_SU,
        choices=constants.PERM_CHOICES,
        verbose_name="Read permission",
        help_text="Default read permissions",
    )
    perm_edit = models.IntegerField(
        default=constants.PERM_SU,
        choices=constants.PERM_CHOICES,
        verbose_name="Edit permission",
        help_text="Default edit permissions",
    )
    markup_engine = models.CharField(
        max_length=255,
        default=app_settings.MARKUP_ENGINE_DEFAULT,
        choices=engine_field_choices(),
    )
    users = models.ManyToManyField(
        get_user_model(), through="WikiPermissions", related_name="wikis"
    )

    class Meta:
        ordering = ("title",)

    def __str__(self):
        return "%s" % self.title

    def get_absolute_url(self):
        return utils.reverse_to_page(
            "powerwiki:page", self.slug, app_settings.FRONT_PATH
        )

    def can_read(self, user):
        """
        Check that the user can read
        Returns bool
        """
        # Check generic wiki permission
        if utils.has_permission(user, self.perm_read):
            return True

        # Check overrides
        if not user.is_authenticated:
            return False
        try:
            user_perm = self.user_permissions.get(user=user)
        except self.user_permissions.model.DoesNotExist:
            return False
        return user_perm.can_read

    def can_edit(self, user):
        """
        Check that the user can read
        Returns bool
        """
        # Check generic wiki permission
        if utils.has_permission(user, self.perm_edit):
            return True

        # Check overrides
        if not user.is_authenticated:
            return False
        try:
            user_perm = self.user_permissions.get(user=user)
        except self.user_permissions.model.DoesNotExist:
            return False
        return user_perm.can_edit

    def gen_breadcrumbs(self, page_path=None):
        """
        Given a wiki and page path, build breadcrumbs
        """
        breadcrumbs = [
            {"title": self.title, "class": "", "url": self.get_absolute_url()}
        ]
        if page_path and page_path != app_settings.FRONT_PATH:
            path_root = ""
            for path_fragment in page_path.split("/"):
                path = path_root + path_fragment
                try:
                    crumb = Page.objects.get(wiki=self, path=path)
                    breadcrumbs.append(
                        {
                            "title": crumb.full_title(),
                            "class": "",
                            "url": crumb.get_absolute_url(),
                        }
                    )
                except Page.DoesNotExist:
                    breadcrumbs.append(
                        {
                            "title": utils.title_from_path(path_fragment),
                            "class": " doesnotexist",
                            "url": utils.reverse_to_page(
                                "powerwiki:page-edit", self.slug, path,
                            ),
                        }
                    )
                path_root += path_fragment + "/"
        return breadcrumbs

    def search(self, query):
        if "django.contrib.postgres" in settings.INSTALLED_APPS:
            return self.pages.annotate(search=SearchVector("title", "content")).filter(
                search=query
            )
        else:
            return self.pages.filter(
                models.Q(title__icontains=query) | models.Q(content__icontains=query)
            )


class WikiPermissions(models.Model):
    user = models.ForeignKey(
        get_user_model(), related_name="wiki_permissions", on_delete=models.CASCADE
    )
    wiki = models.ForeignKey(
        Wiki, related_name="user_permissions", on_delete=models.CASCADE
    )
    can_read = models.BooleanField(default=True, help_text="User can read the wiki")
    can_edit = models.BooleanField(default=True, help_text="User can edit the wiki")


class Page(models.Model):
    """
    Wiki page
    """

    wiki = models.ForeignKey(Wiki, related_name="pages", on_delete=models.CASCADE)
    title = models.CharField(max_length=255, help_text="Title of the page",)
    path = models.CharField(
        max_length=100,
        validators=[
            RegexValidator(
                r"^[a-z0-9][a-z0-9_\-]*(/[a-z0-9][a-z0-9_\-]*)*$",
                (
                    "A path can only be lowercase, numbers, hyphens and underscores, "
                    "using forward slashes to separate sub-pages, "
                    "with each page starting with a lowercase or number."
                ),
                "invalid",
            )
        ],
        help_text="Path for the page, under the wiki root",
    )
    is_locked = models.BooleanField(
        default=False, help_text="If locked, can only be edited by wiki admin",
    )
    content = models.TextField(blank=True, help_text="Page content")
    markup_engine = models.CharField(
        max_length=255,
        default=app_settings.MARKUP_ENGINE_DEFAULT,
        choices=engine_field_choices(),
    )

    class Meta:
        unique_together = ("wiki", "path")
        ordering = ("title",)

    def __str__(self):
        return self.title

    def get_absolute_url(self):
        return utils.reverse_to_page("powerwiki:page", self.wiki.slug, self.path)

    def get_edit_url(self):
        return utils.reverse_to_page("powerwiki:page-edit", self.wiki.slug, self.path)

    def full_title(self):
        return "%s" % (self.title)

    def render_content(self, user: AbstractUser):
        engine_class = get_engine(self.markup_engine)
        engine = engine_class(wiki=self.wiki, page=self)
        html = engine.render(self.content, user=user)
        return mark_safe(html)


def asset_upload_to(asset, filename):
    return os.path.join("powerwiki", asset.wiki.slug, filename)


class Asset(models.Model):
    """
    Wiki asset
    """

    wiki = models.ForeignKey(Wiki, related_name="assets", on_delete=models.CASCADE)
    name = models.SlugField(help_text="Internal name for the asset")
    file = models.FileField(upload_to=asset_upload_to)

    def get_absolute_url(self):
        """
        Asset download url
        """
        return utils.reverse_to_asset("powerwiki:asset", self.wiki.slug, self.name)

    def get_edit_url(self):
        return utils.reverse_to_asset("powerwiki:asset-edit", self.wiki.slug, self.name)

    def get_media_url(self):
        return self.file.url



from django.contrib import messages
from django.http import HttpResponseRedirect
from django.shortcuts import get_object_or_404, redirect, render

from . import app_settings, archive, forms, models
from .decorators import edit_required, get_wiki, permission_required, read_required
from .utils import reverse_to_asset, reverse_to_page, title_from_path


@permission_required(app_settings.PERM_INDEX)
def index(request):
    """
    Index redirects to /index/
    """
    # Get all wikis that the user is allowed to read
    wikis = []
    for wiki in models.Wiki.objects.all():
        if wiki.can_read(request.user):
            wikis.append(wiki)

    return render(
        request,
        "powerwiki/index.html",
        {"title": "Available Wikis", "wikis": wikis, "body_class": "powerwiki_index"},
    )


@get_wiki
@read_required
def page(request, wiki, wiki_slug, page_path=None):
    """
    Show a wiki page
    """
    # Enforce canonical paths
    if page_path and page_path != page_path.lower():
        return HttpResponseRedirect(
            reverse_to_page("powerwiki:page", wiki_slug, page_path.lower())
        )

    # Hide the page path
    if not page_path:
        page_path = app_settings.FRONT_PATH

    # Look up page
    try:
        page = models.Page.objects.get(wiki=wiki, path=page_path)
        title = page.full_title()
    except models.Page.DoesNotExist:
        page = None
        title = "New page: %s" % title_from_path(page_path)

    if wiki.can_edit(request.user):
        edit_url = reverse_to_page("powerwiki:page-edit", wiki_slug, page_path)

    else:
        edit_url = None

    # Prepare page content
    return render(
        request,
        "powerwiki/show.html",
        {
            "page": page,
            "title": title,
            "wiki_slug": wiki_slug,
            "breadcrumbs": wiki.gen_breadcrumbs(page_path),
            "edit_url": edit_url,
            "body_class": "powerwiki_show",
        },
        status=404 if not page else None,
    )


@get_wiki
@read_required
def search(request, wiki, wiki_slug):
    """
    Search a wiki
    """
    query = request.GET.get("q", "").strip()
    pages = wiki.search(query)

    return render(
        request,
        "powerwiki/search.html",
        {
            "title": "Search",
            "wiki_slug": wiki_slug,
            "breadcrumbs": wiki.gen_breadcrumbs(app_settings.FRONT_PATH)
            + [{"title": "Search", "class": "", "url": ""}],
            "search_query": query,
            "pages": pages,
            "body_class": "powerwiki_search",
        },
    )


@get_wiki
@edit_required
def page_edit(request, wiki, wiki_slug, page_path):
    """
    Edit a wiki page
    """
    # Page paths must always be lower case
    if page_path != page_path.lower():
        return HttpResponseRedirect(
            reverse_to_page("powerwiki:page-edit", wiki_slug, page_path.lower())
        )

    # Look up page
    try:
        page = models.Page.objects.get(wiki=wiki, path=page_path)
    except models.Page.DoesNotExist:
        page = None

    # Save or display form
    if request.method == "POST":
        if page:
            form = forms.PageForm(request.POST, instance=page)
        else:
            form = forms.PageForm(request.POST)

        if form.is_valid():
            # Check wiki hasn't been modified in the form
            if form.cleaned_data["wiki"] != wiki:
                # Don't worry about being helpful - this is a deliberate
                # attempt to subvert the authentication system
                messages.error(request, "Trying to save to invalid wiki.")
            elif not form.cleaned_data["content"].strip():
                page.delete()
                messages.success(request, "Page deleted.")
                return HttpResponseRedirect(
                    reverse_to_page("powerwiki:page", wiki_slug)
                )
            else:
                form.save()
                messages.success(request, "Page saved.")
                return HttpResponseRedirect(
                    reverse_to_page("powerwiki:page", wiki_slug, page_path)
                )
        else:
            messages.error(request, "Error processing form.")

    else:
        if page:
            form = forms.PageForm(instance=page)
        else:
            form = forms.PageForm(
                initial={
                    "wiki": wiki,
                    "title": title_from_path(page_path),
                    "path": page_path,
                }
            )

    return render(
        request,
        "powerwiki/edit.html",
        {
            "form": form,
            "title": "Edit page: %s" % page_path,
            "breadcrumbs": wiki.gen_breadcrumbs(page_path) + [{"title": "Edit page"}],
            "show_url": reverse_to_page("powerwiki:page", wiki_slug, page_path),
            "body_class": "powerwiki_edit",
        },
    )


@get_wiki
@edit_required
def wiki_import(request, wiki, wiki_slug):
    """
    Import a wiki
    """
    if request.method == "POST":
        form = forms.ImportForm(request.POST, request.FILES)
        if form.is_valid():
            # Wipe
            if form.cleaned_data["wipe"]:
                wiki.pages.all().delete()

            # Import
            try:
                archive.import_zip(wiki, request.FILES["file"])
            except archive.ZipImportError as e:
                messages.error(request, "Wiki could not be imported: %s" % e)
            else:
                messages.success(request, "Wiki imported")
                return HttpResponseRedirect(
                    reverse_to_page("powerwiki:page", wiki_slug)
                )
    else:
        form = forms.ImportForm()

    return render(
        request,
        "powerwiki/import.html",
        {
            "form": form,
            "title": "Import wiki: %s" % wiki_slug,
            "show_url": reverse_to_page("powerwiki:page", wiki_slug),
            "body_class": "powerwiki_import",
        },
    )


@get_wiki
@read_required
def asset(request, wiki, wiki_slug, asset_name):
    """
    Show asset details
    """
    # Asset slugs must always be lower case
    if asset_name != asset_name.lower():
        return HttpResponseRedirect(
            reverse_to_asset("powerwiki:asset", wiki_slug, asset_name.lower())
        )

    # Get asset
    asset = get_object_or_404(wiki.assets, name=asset_name)

    return redirect(asset.get_media_url())


@get_wiki
@edit_required
def asset_edit(request, wiki, wiki_slug, asset_name):
    """
    Edit an asset
    """
    # Asset slugs must always be lower case
    if asset_name != asset_name.lower():
        return HttpResponseRedirect(
            reverse_to_asset("powerwiki:asset", wiki_slug, asset_name.lower())
        )

    # Look up asset
    initial = {}
    try:
        asset = models.Asset.objects.get(wiki=wiki, name=asset_name)
    except models.Asset.DoesNotExist:
        asset = None
        initial = {
            "wiki": wiki,
            "name": asset_name,
        }

    # Save or display form
    if request.method == "POST":
        form = forms.AssetForm(request.POST, request.FILES, instance=asset)

        if form.is_valid():
            # Check wiki hasn't been modified in the form
            if form.cleaned_data["wiki"] != wiki:
                # Don't worry about being helpful - this is a deliberate
                # attempt to subvert the authentication system
                messages.error(request, "Trying to save to invalid wiki.")
            else:
                asset = form.save()
                messages.success(request, "Asset saved.")
                return HttpResponseRedirect(
                    reverse_to_asset("powerwiki:asset", wiki_slug, asset_name)
                )
        else:
            messages.error(request, "Error processing form.")

    else:
        form = forms.AssetForm(initial=initial, instance=asset)

    breadcrumbs = wiki.gen_breadcrumbs()
    if asset:
        breadcrumbs += [{"title": "Asset", "url": asset.get_absolute_url()}]
    breadcrumbs += [{"title": "Edit asset"}]

    return render(
        request,
        "powerwiki/asset_edit.html",
        {
            "form": form,
            "title": "Edit asset %s" % asset_name,
            "breadcrumbs": breadcrumbs,
            "body_class": "powerwiki_asset_edit",
        },
    )



"""
Test parse_url in html.py
"""
from django.core.exceptions import ValidationError

import pytest

from powerwiki.constants import SCHEME_ASSET, SCHEME_WIKI
from powerwiki.html import parse_url


def test_parse_url__absolute_url__unchanged(wiki, page):
    url = "https://example.com"
    parsed = parse_url(url, wiki, page)
    assert parsed is None


def test_parse_url__mailto__unchanged(wiki, page):
    url = "mailto:bob@example.com"
    parsed = parse_url(url, wiki, page)
    assert parsed is None


def test_parse_url__wiki_scheme_as_single__unchanged(settings, wiki, page):
    settings.POWERWIKI_SINGLE_MODE = True
    url = "wiki:page"
    parsed = parse_url(url, wiki, page)
    assert parsed == (SCHEME_WIKI, wiki.slug, "page")


def test_parse_url__wiki_scheme_with_slug_as_single__validation_fails(
    settings, wiki, page
):
    settings.POWERWIKI_SINGLE_MODE = True
    url = "wiki:slug:page"
    with pytest.raises(ValidationError) as excinfo:
        parse_url(url, wiki, page)
    assert (
        excinfo.value.message
        == "Wiki link wiki:slug:page specifies a wiki in single mode"
    )


def test_parse_url__wiki_scheme_with_multiple_colons__validation_fails(wiki, page):
    url = "wiki:slug:page:extra"
    with pytest.raises(ValidationError) as excinfo:
        parse_url(url, wiki, page)
    assert (
        excinfo.value.message == "Wiki link wiki:slug:page:extra has too many sections"
    )


def test_parse_url__wiki_scheme_with_invalid_wiki_slug__validation_fails(wiki, page):
    url = "wiki:invalid#fail:page"
    with pytest.raises(ValidationError) as excinfo:
        parse_url(url, wiki, page)
    assert excinfo.value.message == f"Wiki link {url} has an invalid wiki slug"


def test_parse_url__asset_scheme_with_invalid_wiki_slug__validation_fails(wiki, page):
    url = "asset:invalid#fail:file"
    with pytest.raises(ValidationError) as excinfo:
        parse_url(url, wiki, page)
    assert excinfo.value.message == f"Asset link {url} has an invalid wiki slug"


def test_parse_url__wiki_scheme_with_invalid_path__validation_fails(wiki, page):
    url = "wiki:slug:#fail"
    with pytest.raises(ValidationError) as excinfo:
        parse_url(url, wiki, page)
    assert excinfo.value.message == f"Wiki link {url} has an invalid page path"


def test_parse_url__asset_scheme_with_invalid_name__validation_fails(wiki, page):
    url = "asset:slug:#fail"
    with pytest.raises(ValidationError) as excinfo:
        parse_url(url, wiki, page)
    assert excinfo.value.message == f"Asset link {url} has an invalid asset name"


def test_parse_url__wiki_scheme__unchanged(wiki, page):
    url = "wiki:page"
    parsed = parse_url(url, wiki, page)
    assert parsed == (SCHEME_WIKI, wiki.slug, "page")


def test_parse_url__wiki_scheme_with_current_slug__unchanged(wiki, page):
    url = "wiki:slug:page"
    parsed = parse_url(url, wiki, page)
    assert parsed == (SCHEME_WIKI, "slug", "page")


def test_parse_url__wiki_scheme_with_other_slug__unchanged(wiki, page):
    url = "wiki:other:page"
    parsed = parse_url(url, wiki, page)
    assert parsed == (SCHEME_WIKI, "other", "page")


def test_parse_url__implied_wiki_scheme__unchanged(wiki, page):
    url = ":page"
    parsed = parse_url(url, wiki, page)
    assert parsed == (SCHEME_WIKI, wiki.slug, "page")


def test_parse_url__implied_wiki_scheme_with_current_slug__unchanged(wiki, page):
    url = ":slug:page"
    parsed = parse_url(url, wiki, page)
    assert parsed == (SCHEME_WIKI, "slug", "page")


def test_parse_url__implied_wiki_scheme_with_other_slug__unchanged(wiki, page):
    url = ":other:page"
    parsed = parse_url(url, wiki, page)
    assert parsed == (SCHEME_WIKI, "other", "page")


def test_parse_url__asset_scheme__unchanged(wiki, page):
    url = "asset:file"
    parsed = parse_url(url, wiki, page)
    assert parsed == (SCHEME_ASSET, wiki.slug, "file")


def test_parse_url__asset_scheme_with_current_slug__unchanged(wiki, page):
    url = "asset:slug:file"
    parsed = parse_url(url, wiki, page)
    assert parsed == (SCHEME_ASSET, "slug", "file")


def test_parse_url__asset_scheme_with_other_slug__unchanged(wiki, page):
    url = "asset:other:file"
    parsed = parse_url(url, wiki, page)
    assert parsed == (SCHEME_ASSET, "other", "file")


def test_parse_url__absolute_outside_wiki__unchanged(wiki, page):
    url = "/other"
    parsed = parse_url(url, wiki, page)
    assert parsed is None



import pytest
from model_bakery import baker

from powerwiki.models import Asset, Page, Wiki


@pytest.fixture
def wiki(db):
    return baker.make(Wiki, slug="slug")


@pytest.fixture
def page(db, wiki):
    return baker.make(Page, wiki=wiki, path="parent/page")


@pytest.fixture
def asset(db, wiki):
    return baker.make(Asset, wiki=wiki, name="image", _create_files=True)



from django.contrib.auth.models import AbstractUser
from django.utils.html import linebreaks

from .base import MarkupEngine


class PlainText(MarkupEngine):
    label = "Plain text"
    extensions = ["extra"]

    def to_html(self, raw: str, user: AbstractUser) -> str:
        return linebreaks(raw, autoescape=True)



from django.contrib import admin

from .models import Asset, Page, Wiki, WikiPermissions


class WikiPermissionsInline(admin.TabularInline):
    model = WikiPermissions
    extra = 2


class WikiAdmin(admin.ModelAdmin):
    list_display = ["title", "slug", "perm_read", "perm_edit"]
    prepopulated_fields = {"slug": ("title",)}
    inlines = [
        WikiPermissionsInline,
    ]


admin.site.register(Wiki, WikiAdmin)


class PageAdmin(admin.ModelAdmin):
    list_display = ["wiki", "title", "path", "is_locked"]
    list_filter = ("wiki", "is_locked")
    search_fields = ["title", "content"]
    prepopulated_fields = {"path": ("title",)}


admin.site.register(Page, PageAdmin)


class AssetAdmin(admin.ModelAdmin):
    list_display = ["name"]


admin.site.register(Asset, AssetAdmin)



from django.urls import re_path

from . import app_settings, views
from .constants import ASSET_NAME_PATTERN, PAGE_PATH_PATTERN, WIKI_SLUG_PATTERN


app_name = "powerwiki"


url_dict = {}
url_prefix = ""

if app_settings.SINGLE_MODE:
    url_dict = {
        "wiki_slug": app_settings.SINGLE_SLUG,
    }

else:
    url_prefix = fr"(?P<wiki_slug>{WIKI_SLUG_PATTERN})/"


urlpatterns = [
    re_path(
        fr"^{url_prefix}(?P<page_path>{PAGE_PATH_PATTERN})/_edit/$",
        views.page_edit,
        url_dict,
        name="page-edit",
    ),
    re_path(
        fr"^{url_prefix}(?P<page_path>{PAGE_PATH_PATTERN})/$",
        views.page,
        url_dict,
        name="page",
    ),
    re_path(fr"^{url_prefix}_import/$", views.wiki_import, url_dict, name="import",),
    re_path(fr"^{url_prefix}_search/$", views.search, url_dict, name="search",),
    re_path(
        fr"^{url_prefix}_asset/(?P<asset_name>{ASSET_NAME_PATTERN})/$",
        views.asset,
        url_dict,
        name="asset",
    ),
    re_path(
        fr"^{url_prefix}_asset/(?P<asset_name>{ASSET_NAME_PATTERN})/_edit/$",
        views.asset_edit,
        url_dict,
        name="asset-edit",
    ),
    re_path(fr"^{url_prefix}$", views.page, url_dict, name="page"),
]

if not app_settings.SINGLE_MODE:
    urlpatterns.append(re_path(r"^$", views.index, name="index"),)



from yaa_settings import AppSettings

from . import constants


class PowerwikiSettings(AppSettings):
    prefix = "POWERWIKI"

    # If True, run powerwiki in single wiki mode
    SINGLE_MODE = False

    # Slug for the wiki definition to use in single wiki mode
    # Must be lower case
    def SINGLE_SLUG(self, value):
        return (value or "default").lower()

    # Permissions for who can see the wiki index (list of wikis)
    # Users will only see those wikis which they have access to
    PERM_INDEX = constants.PERM_PUBLIC

    # Path for the front page of the wiki
    # Must be lower case
    def FRONT_PATH(self, value):
        return (value or "index").lower()

    # Tags to check for wiki links
    # List of (tag, attr) tuples
    #   tag     Name of HTML tag
    #   attr    Field which contains the wiki link
    LINK_TAGS = [
        ("a", "href"),
        ("img", "src"),
    ]

    # List of LINK_TAGS tags which should have their contents if blank
    # If the target path exists in the database, the title will be used
    # If it doesn't, it will use the path
    LINK_TAG_CONTAINERS = ["a"]

    # Available markup engines
    MARKUP_ENGINES = [
        "powerwiki.markup.rest.RestructuredText",
        "powerwiki.markup.md.Markdown",
        "powerwiki.markup.plain.PlainText",
    ]

    # Default markup engine
    MARKUP_ENGINE_DEFAULT = "powerwiki.markup.rest.RestructuredText"

    # HTML parser for BeautifulSoup
    HTML_PARSER = "html.parser"



from django import template
from django.utils.safestring import mark_safe


register = template.Library()


@register.simple_tag(takes_context=True)
def render_wiki(context, page):
    rendered = page.render_content(user=context["request"].user)
    return mark_safe(rendered)



from django.contrib.auth.models import AbstractUser

from docutils import nodes, utils
from docutils.core import publish_parts
from docutils.parsers.rst.roles import _roles, set_implicit_options

from ..utils import wikipathify
from .base import MarkupEngine


def role_link_wiki(name, rawtext, text, lineno, inliner, options={}, content=[]):
    node = nodes.reference(
        rawtext, utils.unescape(text), refuri=wikipathify(text), **options
    )
    return [node], []


def role_link_asset(name, rawtext, text, lineno, inliner, options={}, content=[]):
    node = nodes.reference(
        rawtext, utils.unescape(text), refuri=f"asset:{wikipathify(text)}", **options
    )
    return [node], []


# According to docutils we're supposed to call roles.register_local_role, but that would
# potentially clash with any other use of docutils in the system. We'll therefore
# reimplement it in this module - calling `set_implicit_options` directly, then setting
# and unsetting _roles within the markup engine later using TmpRole
set_implicit_options(role_link_wiki)
set_implicit_options(role_link_asset)


class TmpRole:
    def __init__(self, name, fn):
        self.name = name
        self.fn = fn

    def __enter__(self):
        self.archive = _roles.pop(self.name, None)
        _roles[self.name] = self.fn

    def __exit__(self, type, value, traceback):
        del _roles[self.name]
        if self.archive is not None:
            _roles[self.name] = self.archive


class RestructuredText(MarkupEngine):
    label = "reStructuredText"
    docutils_settings = {
        "doctitle_xform": False,
    }

    def to_html(self, raw: str, user: AbstractUser) -> str:
        # Prepare docutils for custom roles
        with TmpRole("wiki", role_link_wiki), TmpRole("asset", role_link_asset):
            # Convert to HTML
            parts = publish_parts(
                source=raw,
                writer_name="html",
                settings_overrides=self.docutils_settings,
            )
        html = parts["fragment"]
        return html



from django.contrib.auth.models import AbstractUser

import markdown

from .base import MarkupEngine


class Markdown(MarkupEngine):
    label = "Markdown"
    extensions = ["extra"]

    def to_html(self, raw: str, user: AbstractUser) -> str:
        md = markdown.markdown(raw, extensions=self.extensions)
        return md



"""
Powerwiki utils
"""
import re

from django.urls import reverse
from django.utils.safestring import mark_safe

from . import app_settings, constants


def has_permission(user, permission):
    """
    See if the user has the specified level of permission
    """
    # If public, all can see
    if permission == constants.PERM_PUBLIC:
        return True

    # If not public, must be PERM_USER or higher
    if not user.is_authenticated:
        return False

    # Superuser can see all
    # Staff can see STAFF and lower
    # Anyone here must be a user, so can see USERS
    if (
        user.is_superuser
        or (user.is_staff and permission <= constants.PERM_STAFF)
        or permission == constants.PERM_USERS
    ):
        return True

    return False


def reverse_to_page(view, wiki_slug, page_path=None):
    """
    Return the URL for the given page
    """
    kwargs = {}
    if not app_settings.SINGLE_MODE:
        kwargs["wiki_slug"] = wiki_slug
    if page_path and not (
        view == "powerwiki:page" and page_path == app_settings.FRONT_PATH
    ):
        kwargs["page_path"] = page_path
    return reverse(view, kwargs=kwargs)


def wikipathify(value):
    """
    Like django.utils.text.slugify, but supports sub-pages
    """
    value = re.sub(r"(^/|[^\w\s/-]|/$)", "", value).strip().lower()
    return mark_safe(re.sub(r"[-\s]+", "-", value))


def title_from_path(value):
    """
    Make a best guess at converting a wiki path into a title
    """
    # Find slug at end of path
    slug_index = value.rfind("/")
    if slug_index > -1:
        value = value[slug_index + 1 :]

    return re.sub(r"[-\s_]+", " ", value).title()


def reverse_to_asset(view, wiki_slug, asset_name=None):
    """
    Return the URL for the given asset view
    """
    kwargs = {}
    if not app_settings.SINGLE_MODE:
        kwargs["wiki_slug"] = wiki_slug
    if asset_name:
        kwargs["asset_name"] = asset_name
    return reverse(view, kwargs=kwargs)



"""
Test conversion functions in html.py
"""

from django.contrib.auth.models import AnonymousUser
from django.test.html import parse_html

from powerwiki.html import process


def test_process__sample__links_are_updated(wiki, page, asset):
    raw = f"""
    <p><a href="/app">example</a></p>
    <p><a href="other">other</a></p>
    <p><a href="wiki:sibling">sibling</a></p>
    <p><img src="asset:{asset.name}"></p>
    """
    expected = f"""
    <p>
        <a href="/app">example</a>
    </p>
    <p>
        <a
            href="/wiki/slug/parent/other/"
            data-edit="/wiki/slug/parent/other/_edit/"
            data-missing="True"
            class="powerwiki__wiki"
        >other</a>
    </p>
    <p>
        <a
            href="/wiki/slug/sibling/"
            data-edit="/wiki/slug/sibling/_edit/"
            data-missing="True"
            class="powerwiki__wiki"
        >sibling</a>
    </p>
    <p>
        <img
            src="{asset.file.url}"
            data-edit="/wiki/slug/_asset/{asset.name}/_edit/"
            class="powerwiki__asset"
        >
    </p>
    """

    processed = process(raw, wiki=wiki, page=page, user=AnonymousUser())
    assert parse_html(processed) == parse_html(expected)



# Generated by Django 2.2.15 on 2020-08-31 01:16

import django.core.validators
import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models

import powerwiki.models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name="Wiki",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "title",
                    models.CharField(help_text="Title of the wiki", max_length=255),
                ),
                ("slug", models.SlugField(help_text="Slug for the wiki", unique=True)),
                (
                    "description",
                    models.TextField(blank=True, help_text="Description of wiki"),
                ),
                (
                    "perm_read",
                    models.IntegerField(
                        choices=[
                            (3, "Superusers only"),
                            (2, "Staff only"),
                            (1, "All users"),
                            (0, "Public"),
                        ],
                        default=3,
                        help_text="Default read permissions",
                        verbose_name="Read permission",
                    ),
                ),
                (
                    "perm_edit",
                    models.IntegerField(
                        choices=[
                            (3, "Superusers only"),
                            (2, "Staff only"),
                            (1, "All users"),
                            (0, "Public"),
                        ],
                        default=3,
                        help_text="Default edit permissions",
                        verbose_name="Edit permission",
                    ),
                ),
                (
                    "markup_engine",
                    models.CharField(
                        choices=[
                            (
                                "powerwiki.markup.rest.RestructuredText",
                                "reStructuredText",
                            ),
                            ("powerwiki.markup.md.Markdown", "Markdown"),
                            ("powerwiki.markup.plain.PlainText", "Plain text"),
                        ],
                        default="powerwiki.markup.rest.RestructuredText",
                        max_length=255,
                    ),
                ),
            ],
            options={"ordering": ("title",),},
        ),
        migrations.CreateModel(
            name="WikiPermissions",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "can_read",
                    models.BooleanField(
                        default=True, help_text="User can read the wiki"
                    ),
                ),
                (
                    "can_edit",
                    models.BooleanField(
                        default=True, help_text="User can edit the wiki"
                    ),
                ),
                (
                    "user",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="wiki_permissions",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                (
                    "wiki",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="user_permissions",
                        to="powerwiki.Wiki",
                    ),
                ),
            ],
        ),
        migrations.AddField(
            model_name="wiki",
            name="users",
            field=models.ManyToManyField(
                related_name="wikis",
                through="powerwiki.WikiPermissions",
                to=settings.AUTH_USER_MODEL,
            ),
        ),
        migrations.CreateModel(
            name="Asset",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.SlugField(help_text="Internal name for the asset")),
                ("file", models.FileField(upload_to=powerwiki.models.asset_upload_to)),
                (
                    "wiki",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="assets",
                        to="powerwiki.Wiki",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="Page",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "title",
                    models.CharField(help_text="Title of the page", max_length=255),
                ),
                (
                    "path",
                    models.CharField(
                        help_text="Path for the page, under the wiki root",
                        max_length=100,
                        validators=[
                            django.core.validators.RegexValidator(
                                "^[a-z0-9][a-z0-9_\\-]*(/[a-z0-9][a-z0-9_\\-]*)*$",
                                "A path can only be lowercase, numbers, hyphens and underscores, using forward slashes to separate sub-pages, with each page starting with a lowercase or number.",
                                "invalid",
                            )
                        ],
                    ),
                ),
                (
                    "is_locked",
                    models.BooleanField(
                        default=False,
                        help_text="If locked, can only be edited by wiki admin",
                    ),
                ),
                ("content", models.TextField(blank=True, help_text="Page content")),
                (
                    "markup_engine",
                    models.CharField(
                        choices=[
                            (
                                "powerwiki.markup.rest.RestructuredText",
                                "reStructuredText",
                            ),
                            ("powerwiki.markup.md.Markdown", "Markdown"),
                            ("powerwiki.markup.plain.PlainText", "Plain text"),
                        ],
                        default="powerwiki.markup.rest.RestructuredText",
                        max_length=255,
                    ),
                ),
                (
                    "wiki",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="pages",
                        to="powerwiki.Wiki",
                    ),
                ),
            ],
            options={"ordering": ("title",), "unique_together": {("wiki", "path")},},
        ),
    ]



"""
Decorators
"""

from django.contrib.auth.decorators import user_passes_test
from django.core.exceptions import PermissionDenied
from django.shortcuts import get_object_or_404

from . import constants, models, utils


#
# Wiki util
#


def get_wiki(fn):
    """
    Look up the wiki in kwargs['wiki_slug']
    Return 404 if not found, otherwise pass on to decorated fn as second arg
    """

    def wrapper(request, *args, **kwargs):
        wiki = get_object_or_404(models.Wiki, slug=kwargs.get("wiki_slug", None))
        return fn(request, wiki, *args, **kwargs)

    return wrapper


#
# Permission handling
#


def permission_required(permission, login_url=None):
    """
    Ensures that the user has specified permission to access the current page
    """

    def check_perms(user):
        # Check if user not logged in
        if not user.is_authenticated:
            if permission == constants.PERM_PUBLIC:
                # They're not, but it's public
                return True

            # User not logged in, needs a permission, show the login form
            return False

        # User is logged in, just lacks permission
        if not utils.has_permission(user, permission):
            raise PermissionDenied

        return True

    return user_passes_test(check_perms, login_url=login_url)


def read_required(fn):
    """
    Must be called on a function already decorated with @get_wiki
    """

    def wrapper(request, wiki, *args, **kwargs):
        if not wiki.can_read(request.user):
            if request.user.is_authenticated:
                raise PermissionDenied
            # Make auth decorator fail to force user to login form
            return user_passes_test(lambda u: False)(lambda r: None)(request)
        return fn(request, wiki, *args, **kwargs)

    return wrapper


def edit_required(fn):
    """
    Must be called on a function already decorated with @get_wiki
    """

    def wrapper(request, wiki, *args, **kwargs):
        if not wiki.can_edit(request.user):
            if request.user.is_authenticated:
                raise PermissionDenied
            # Make auth decorator fail to force user to login form
            return user_passes_test(lambda u: False)(lambda r: None)(request)
        return fn(request, wiki, *args, **kwargs)

    return wrapper



from django.utils.module_loading import import_string

from .. import app_settings


engines = {}


def load_engines():
    for class_name in app_settings.MARKUP_ENGINES:
        engines[class_name] = import_string(class_name)
    return engines


def get_engine(class_name):
    # Just sanity check to make sure something odd isn't going on
    if class_name not in engines:
        raise ValueError(f"Unknown wiki markup engine: {class_name}")
    return engines[class_name]


def engine_field_choices():
    return [
        (engine_name, engines[engine_name].label)
        for engine_name in app_settings.MARKUP_ENGINES
    ]



"""
Powerwiki constants
"""

# Permissions
PERM_SU = 3
PERM_STAFF = 2
PERM_USERS = 1
PERM_PUBLIC = 0
PERM_CHOICES = (
    (PERM_SU, "Superusers only"),
    (PERM_STAFF, "Staff only"),
    (PERM_USERS, "All users"),
    (PERM_PUBLIC, "Public"),
)

# URL scheme names
SCHEME_WIKI = "wiki"
SCHEME_ASSET = "asset"
SCHEME_IMAGE = "image"

# URL patterns
WIKI_SLUG_PATTERN = r"[a-zA-Z0-9][-\w]*"
PAGE_PATH_PATTERN = r"[a-zA-Z0-9].*?(/[a-zA-Z0-9][- \w]*)*"
ASSET_NAME_PATTERN = r"[a-zA-Z0-9].*?"

# Wikilink tags
WIKILINK_OPEN = "[["
WIKILINK_SCHEME_SEPARATOR = ":"
WIKILINK_LABEL_SEPARATOR = "|"
WIKILINK_CLOSE = "]]"



"""
Wiki archive management
"""
import os
import zipfile

from django.core.files.uploadedfile import InMemoryUploadedFile

from . import models
from .utils import title_from_path


TITLE_PREFIX = "TITLE: "


class ZipImportError(Exception):
    pass


def import_zip(wiki, file):
    """
    Import a zip into a wiki
    Zip should contain two directories:
        pages
            page.txt            # path: page
            page
                subpage.txt     # path: page/subpage
        assets
            internal_id.jpg

    Page extensions must be .txt or .md - both are treated as pretext format
    Asset extensions should indicate filetype.

    Page titles are expected on the first line, with the prefix "TITLE: "; if
    the first line does not match this format, it will be considered to be part
    of the content, and the title will be based on the path.

    These can optionally be held inside a directory whose name matches the zip,
    for example:

        mywiki.zip
            mywiki
                pages
                assets
    """
    # Test filename
    zipname, ext = os.path.splitext(file.name)
    if (
        not isinstance(file, InMemoryUploadedFile)
        and not zipfile.is_zipfile(file.temporary_file_path)
    ) or ext.lower() != ".zip":
        raise ZipImportError("File is not a zip file")

    # Open the zip file
    try:
        zf = zipfile.ZipFile(file)
    except zipfile.BadZipFile as e:
        raise ZipImportError("Bad zip file: %s" % e)
    except zipfile.LargeZipFile:
        raise ZipImportError("Bad zip file: ZIP64 not enabled")
    filenames = [info.filename for info in zf.infolist()]

    # Check the zip for optional root container
    path_con = zipname + "/"
    path_pages = "pages/"
    path_assets = "assets/"
    if path_con in filenames:
        path_pages = path_con + path_pages
        path_assets = path_con + path_assets

    # Check the structure of the zip
    if path_pages not in filenames:
        raise ZipImportError("Pages directory not found")
    if path_assets not in filenames:
        raise ZipImportError("Assets directory not found")

    # Import the pages
    pages = [filename for filename in filenames if filename.startswith(path_pages)]
    for filename in pages:
        # Find page path and read content
        page_path, ext = os.path.splitext(filename[len(path_pages) :])
        if ext.lower() not in [".txt", ".pt"]:
            continue
        content = zf.read(filename)

        # Get or create Page
        try:
            page = models.Page.objects.get(wiki=wiki, path=page_path)
        except models.Page.DoesNotExist:
            page = models.Page(wiki=wiki, path=page_path)

        # Set title and content
        lines = content.splitlines()
        if lines and lines[0].startswith(TITLE_PREFIX):
            page.title = lines.pop(0)[TITLE_PREFIX:].strip()
        else:
            page.title = title_from_path(page_path)
        page.content = "\n".join(lines)
        page.save()


def export_zip(wiki):
    """
    Export a wiki into a zip, returning the file object
    """
    raise NotImplementedError()



from django import forms

from .models import Asset, Page


class PageForm(forms.ModelForm):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # Inherit markup engine from parent
        if not self.instance.pk and self.initial.get("wiki"):
            self.initial["markup_engine"] = self.initial["wiki"].markup_engine

    class Meta:
        model = Page
        fields = ["wiki", "path", "title", "markup_engine", "content"]
        widgets = {
            "wiki": forms.HiddenInput(),
            "path": forms.HiddenInput(),
            "title": forms.TextInput(
                attrs={"placeholder": "Title", "autocomplete": "off"}
            ),
            "content": forms.Textarea(),
        }


class ImportForm(forms.Form):
    file = forms.FileField(help_text="Zip containing wiki pages")
    wipe = forms.BooleanField(
        required=False, initial=False, help_text="Wipe the wiki first",
    )


class AssetForm(forms.ModelForm):
    class Meta:
        model = Asset
        fields = ["wiki", "name", "file"]
        widgets = {
            "wiki": forms.HiddenInput(),
            "name": forms.HiddenInput(),
        }



"""
HTML tools
"""
import re
from collections import defaultdict
from functools import lru_cache
from typing import Optional, Tuple
from urllib.parse import urljoin, urlparse

from django.contrib.auth.models import AbstractUser
from django.core.exceptions import ValidationError
from django.template.defaultfilters import slugify
from django.urls import reverse

from bs4 import BeautifulSoup

from . import app_settings
from .constants import (
    ASSET_NAME_PATTERN,
    PAGE_PATH_PATTERN,
    SCHEME_ASSET,
    SCHEME_IMAGE,
    SCHEME_WIKI,
    WIKI_SLUG_PATTERN,
    WIKILINK_CLOSE,
    WIKILINK_LABEL_SEPARATOR,
    WIKILINK_OPEN,
    WIKILINK_SCHEME_SEPARATOR,
)
from .models import Asset, Page, Wiki


re_valid_wiki_slug = re.compile(fr"^{WIKI_SLUG_PATTERN}$")
re_valid_page_path = re.compile(fr"^{PAGE_PATH_PATTERN}$")
re_valid_asset_name = re.compile(fr"^{ASSET_NAME_PATTERN}$")


@lru_cache()
def get_wiki_root():
    return reverse("powerwiki:index")


def parse_url(url: str, wiki: Wiki, page: Page) -> Optional[Tuple[str, str, str]]:
    """
    Parse a wiki URL

    Returns:
        (scheme, wiki_slug, page_path) if it's a wiki url
        None if it's not a wiki url
    """
    # Get the scheme
    if ":" not in url:
        return None

    scheme, wiki_path = url.split(":", 1)
    if scheme == "":
        scheme = SCHEME_WIKI
    elif scheme not in [SCHEME_WIKI, SCHEME_ASSET]:
        return None

    # Resolve wiki
    wiki_slug = wiki.slug
    if ":" in wiki_path:
        # Wiki is specified
        if app_settings.SINGLE_MODE:
            raise ValidationError(f"Wiki link {url} specifies a wiki in single mode")

        wiki_slug, page_path = wiki_path.split(":", 1)
        page_path = page_path.lstrip("/")

        if ":" in page_path:
            raise ValidationError(f"Wiki link {url} has too many sections")

    else:
        wiki_slug = wiki.slug
        page_path = wiki_path

    # Check values are valid
    if not re_valid_wiki_slug.match(wiki_slug):
        if scheme == SCHEME_WIKI:
            raise ValidationError(f"Wiki link {url} has an invalid wiki slug")
        else:
            raise ValidationError(f"Asset link {url} has an invalid wiki slug")
    if scheme == SCHEME_WIKI and not re_valid_page_path.match(page_path):
        raise ValidationError(f"Wiki link {url} has an invalid page path")
    if scheme == SCHEME_ASSET and not re_valid_asset_name.match(page_path):
        raise ValidationError(f"Asset link {url} has an invalid asset name")

    # Ensure path is slugified
    page_path = "/".join([slugify(slug) for slug in page_path.split("/") if slug])

    return scheme, wiki_slug, page_path


def find_wiki_url(url: str, wiki: Wiki, page: Page) -> Optional[Tuple[str, str, str]]:
    """
    Normalises urls into a wiki link, or return None if not a wiki link

    Returns:
        (scheme: str, wiki_slug: str, page_path: str)  A wiki link
    """
    # Absolute urls stay the same
    # TODO: Is there a faster way to do this? This should also detect the current site.
    is_absolute = bool(urlparse(url).netloc)
    if is_absolute:
        return

    # See if we have a wiki or asset url
    try:
        parsed = parse_url(url, wiki, page)
    except ValidationError:
        # Silence errors
        return

    if parsed:
        return parsed

    # Convert relative wiki paths
    page_url = page.get_absolute_url()
    if not url.startswith("./"):
        # Strip trailing / so foo:bar is a sibling of the current page
        # If url starts ./ we want it to be a child
        page_url = page_url.rstrip("/")

    normalised = urljoin(page_url, url)
    wiki_root = get_wiki_root()
    if normalised.startswith(wiki_root):
        # Relative to the root
        relative = normalised[len(wiki_root) :]

        # Slugify the slugs
        slugs = relative.split("/")
        slugs = [slugify(slug) for slug in slugs]
        relative = "/".join(slugs)

        if app_settings.SINGLE_MODE:
            return SCHEME_WIKI, wiki.slug, relative

        elif "/" in relative:
            wiki_slug, page_path = relative.split("/", 1)
            return SCHEME_WIKI, wiki_slug, page_path
        else:
            return SCHEME_WIKI, wiki.slug, relative

    return


def clean(html: str, wiki: Wiki, page: Page):
    """
    Clean new HTML from the user

    * Convert relative paths to wiki urls
    """
    soup = BeautifulSoup(html, features=app_settings.HTML_PARSER)

    for tag_name, attr in app_settings.LINK_TAGS:
        for tag in soup.findAll(tag_name):
            parse_url(tag[attr], wiki, page)


def process(html: str, wiki: Wiki, page: Page, user: AbstractUser):
    """
    Convert wiki links into relative paths
    """
    soup = BeautifulSoup(html, features=app_settings.HTML_PARSER)

    # Convert wiki format links [[page|label]]
    for text in soup.findAll(text=True):
        ptr = 0
        replacement = BeautifulSoup(features=app_settings.HTML_PARSER)
        while ptr > -1:
            # Find next wikilink
            match_open = text.find(WIKILINK_OPEN, ptr)
            if match_open == -1:
                break
            match_close = text.find(WIKILINK_CLOSE, match_open)
            if match_close == -1:
                break

            # Capture text before tag open, and advance ptr to end of area of interest
            if ptr != match_open:
                replacement.append(text[ptr:match_open])
            ptr = match_close + len(WIKILINK_CLOSE)

            # Detect scheme
            path = text[match_open + len(WIKILINK_OPEN) : match_close]
            if WIKILINK_SCHEME_SEPARATOR in path:
                scheme, path = path.split(WIKILINK_SCHEME_SEPARATOR, 1)
            else:
                scheme = SCHEME_WIKI

            # Detect label
            if WIKILINK_LABEL_SEPARATOR in path:
                path, label = path.split(WIKILINK_LABEL_SEPARATOR, 1)
            else:
                label = ""

            # Image scheme can't have a label
            if scheme == SCHEME_IMAGE and label:
                # Invalid tag, skip
                continue

            # Build replacement tag
            if scheme == SCHEME_IMAGE:
                replacement.append(soup.new_tag("img", src=f"{SCHEME_ASSET}:{path}"))
            else:
                link = soup.new_tag("a", href=f"{scheme}:{path}")
                link.string = label
                replacement.append(link)

            # Restart loop to look for next tag

        # Capture anything at the end of the string
        if ptr < len(text):
            replacement.append(text[ptr:])

        # TODO: Might be worth performance testing the above against 3 regexps:
        """
        replaced = re_wiki_link.sub(
            lambda match: '<a href="{path}">{label}</a>'.format(
                path=match.group("path"),
                label=match.group("label") or match.group("path"),
            ),
            text,
        )
        """
        text.replaceWith(replacement)

    # Build lookup tables for tags
    #   found[wiki][path] = [tag, tag, tag ...]
    page_tags = defaultdict(lambda: defaultdict(list))
    asset_tags = defaultdict(lambda: defaultdict(list))

    for tag_name, attr in app_settings.LINK_TAGS:
        for tag in soup.findAll(tag_name):
            wiki_url = find_wiki_url(tag[attr], wiki, page)

            if wiki_url is None:
                continue

            # We need to look up the pages and assets
            # Defer so we can do the lookup in one go
            scheme, wiki_slug, page_slug = wiki_url
            if scheme == SCHEME_WIKI:
                # We don't need to know anything concrete about a link yet, fake it
                page_tags[wiki_slug][page_slug].append((tag, attr))

            elif scheme == SCHEME_ASSET:
                asset_tags[wiki_slug][page_slug].append((tag, attr))

            else:
                continue

            tag["class"] = f"{tag.get('class', '')} powerwiki__{scheme}".strip()

    # Look up wikis and check permissions
    wiki_slugs = list(page_tags.keys()) + list(asset_tags.keys())
    wikis = Wiki.objects.filter(slug__in=wiki_slugs)
    available_wikis = [wiki.slug for wiki in wikis if wiki.can_read(user)]

    # Look up and replace pages
    for wiki_slug, paths in page_tags.items():
        if wiki_slug in available_wikis:
            # Lookup pages for this wiki
            pages = Page.objects.filter(wiki__slug=wiki_slug, path__in=paths.keys())
        else:
            pages = []

        path_to_page = {page.path: page for page in pages}
        fake_wiki = Wiki(slug=wiki_slug)

        # Loop over paths
        for path, tags in paths.items():
            page = path_to_page.get(path)
            if not page:
                page = Page(wiki=fake_wiki, path=path)

            for tag, attr in tags:
                tag[attr] = page.get_absolute_url()
                tag["data-edit"] = page.get_edit_url()
                if not page.pk:
                    tag["data-missing"] = True

                if tag.string == "" and tag.name in app_settings.LINK_TAG_CONTAINERS:
                    if page.pk:
                        tag.string = page.title
                    else:
                        tag.string = path

    # Look up and replace assets
    for wiki_slug, paths in asset_tags.items():
        fake_wiki = Wiki(slug=wiki_slug)

        # Lookup pages for this wiki
        assets = Asset.objects.filter(wiki__slug=wiki_slug, name__in=paths.keys())
        name_to_asset = {asset.name: asset for asset in assets}

        # Loop over paths
        for path, tags in paths.items():
            if path in name_to_asset:
                asset = name_to_asset[path]
                media_url = asset.get_media_url()
            else:
                asset = Asset(wiki=fake_wiki, name=path)
                media_url = ""

            edit_url = asset.get_edit_url()
            for tag, attr in tags:
                tag[attr] = media_url
                tag["data-edit"] = edit_url
                if not asset.pk:
                    tag["data-missing"] = True

    return str(soup)



"""
Test conversion functions in html.py
"""
from model_bakery import baker

from powerwiki.constants import SCHEME_WIKI
from powerwiki.html import find_wiki_url
from powerwiki.models import Page


def test_find_wiki_url__absolute_url__unchanged(wiki, page):
    url = "https://example.com"
    parsed = find_wiki_url(url, wiki, page)
    assert parsed is None


def test_find_wiki_url__mailto__unchanged(wiki, page):
    url = "mailto:bob@example.com"
    parsed = find_wiki_url(url, wiki, page)
    assert parsed is None


def test_find_wiki_url__wiki_scheme_as_single__unchanged(settings, wiki, page):
    settings.POWERWIKI_SINGLE_MODE = True
    url = "wiki:page"
    parsed = find_wiki_url(url, wiki, page)
    assert parsed == (SCHEME_WIKI, wiki.slug, "page")


def test_find_wiki_url__wiki_scheme_with_slug_as_single__validation_fails(
    settings, wiki, page
):
    settings.POWERWIKI_SINGLE_MODE = True
    url = "wiki:slug:page"
    parsed = find_wiki_url(url, wiki, page)
    assert parsed is None


def test_find_wiki_url__relative_outside_wiki__unchanged(wiki, page):
    # current url: /wiki/slug/parent/page/
    # url as absolute: /other/
    url = "../../../other"
    parsed = find_wiki_url(url, wiki, page)
    assert parsed is None


def test_find_wiki_url__absolute_inside_wiki__converted_to_wiki_scheme(wiki, page):
    url = "/wiki/slug/other"
    parsed = find_wiki_url(url, wiki, page)
    assert parsed == (SCHEME_WIKI, "slug", "other")


def test_find_wiki_url__relative_at_wiki_root__converted_to_wiki_scheme(wiki):
    url = "other"
    page = baker.make(Page, wiki=wiki, path="page")
    parsed = find_wiki_url(url, wiki, page)
    assert parsed == (SCHEME_WIKI, "slug", "other")


def test_find_wiki_url__relative_inside_wiki__converted_to_wiki_scheme(wiki, page):
    url = "other"
    parsed = find_wiki_url(url, wiki, page)
    assert parsed == (SCHEME_WIKI, "slug", "parent/other")


def test_find_wiki_url__absolute_child__converted_to_wiki_scheme(wiki, page):
    url = "/wiki/slug/parent/other"
    parsed = find_wiki_url(url, wiki, page)
    assert parsed == (SCHEME_WIKI, "slug", "parent/other")


def test_find_wiki_url__relative_child__converted_to_wiki_scheme(wiki, page):
    url = "./other"
    parsed = find_wiki_url(url, wiki, page)
    assert parsed == (SCHEME_WIKI, "slug", "parent/page/other")


def test_find_wiki_url__relative_parent_sibling__converted_to_wiki_scheme(wiki, page):
    url = "../parent-sibling"
    parsed = find_wiki_url(url, wiki, page)
    assert parsed == (SCHEME_WIKI, "slug", "parent-sibling")


def test_find_wiki_url__absolute_in_different_wiki__converted_to_wiki_scheme(
    wiki, page
):
    url = "/wiki/other/page"
    parsed = find_wiki_url(url, wiki, page)
    assert parsed == (SCHEME_WIKI, "other", "page")


def test_find_wiki_url__relative_in_different_wiki__converted_to_wiki_scheme(
    wiki, page
):
    url = "../../other/page"
    parsed = find_wiki_url(url, wiki, page)
    assert parsed == (SCHEME_WIKI, "other", "page")



from __future__ import annotations

import typing

from django.contrib.auth.models import AbstractUser


if typing.TYPE_CHECKING:
    from ..models import Page, Wiki


class MarkupEngine:
    wiki: Wiki
    page: Page
    label: str = "Unknown"

    def __init__(self, wiki: Wiki, page: Page):
        self.wiki = wiki
        self.page = page

    def render(self, raw: str, user: AbstractUser) -> str:
        raw = self.pre(raw, user)
        html = self.to_html(raw, user)
        html = self.post(html, user)
        return html

    def pre(self, raw: str, user: AbstractUser) -> str:
        return raw

    def to_html(self, raw: str, user: AbstractUser) -> str:
        raise NotImplementedError()

    def post(self, html: str, user: AbstractUser) -> str:
        from ..html import process

        html = process(html, wiki=self.wiki, page=self.page, user=user)
        return html



import re
from pathlib import Path

from setuptools import setup


def find_version(*paths):
    path = Path(*paths)
    content = path.read_text()
    match = re.search(r"^__version__\s*=\s*['\"]([^'\"]*)['\"]", content, re.M)
    if match:
        return match.group(1)
    raise RuntimeError("Unable to find version string.")


setup(version=find_version("powerwiki", "__init__.py"))
